import { type SceneState, type SceneType } from "./base"
import type { Game } from "../game"
import type { SuggestedInteraction } from "../../shared/mutations"
import type { InteractionSceneDefinition } from "../../shared/script/types"
import { getSuggestedInteractionId, type InteractionSceneEvent } from "../../shared/scene-types/interaction"

export const InteractionSceneType = {
    id: "interaction",
    createState(game: Game, definition: InteractionSceneDefinition): InteractionSceneState {
        return new InteractionSceneState(game, definition)
    }
} as const satisfies SceneType<InteractionSceneDefinition, InteractionSceneState>

export class InteractionSceneState implements SceneState<InteractionSceneEvent> {
    private interactionQueue: Map<string, { interaction: SuggestedInteraction; votes: number }> = new Map()
    private ongoingInteractionExecution: SuggestedInteraction | null = null
    private objectVisibilityById = new Map<string, boolean>()

    constructor(private game: Game, private definition: InteractionSceneDefinition) {
        Object.values(definition.objects).forEach(o => this.objectVisibilityById.set(o.id, o.reveal))
    }

    getConnectionEvents(): InteractionSceneEvent[] {
        const events: InteractionSceneEvent[] = []
        events.push(this.getVotesChangedEvent())

        this.objectVisibilityById.entries().forEach(([id, isVisible]) => events.push({ type: "object-visibility-changed", id, isVisible }))

        if (this.ongoingInteractionExecution !== null) events.push({ type: "interaction-execution-started", interaction: this.ongoingInteractionExecution })

        return events
    }

    private emit(event: InteractionSceneEvent) {
        this.game.emit("player-broadcast", { type: "scene-event", event })
    }

    private getVotesChangedEvent() {
        return { type: "votes-changed", votesByInteractionId: Object.fromEntries(this.interactionQueue.entries().map(([id, { votes }]) => ([id, votes]))) } as const
    }

    private emitVotesChanged() {
        this.emit(this.getVotesChangedEvent())
    }

    addInteractionVote(interaction: SuggestedInteraction) {
        const id = getSuggestedInteractionId(interaction)
        const existingItem = this.interactionQueue.get(id)
        if (existingItem === undefined) {
            const item = {
                id,
                votes: 1,
                interaction
            }

            this.interactionQueue.set(item.id, item)
            this.emit({ type: "interaction-queued", interaction })
        } else {
            existingItem.votes += 1
            this.emitVotesChanged()
        }
    }

    removeInteractionVote(id: string) {
        const item = this.interactionQueue.get(id)
        if (item !== undefined) {
            item.votes -= 1
            if (item.votes <= 0) this.interactionQueue.delete(id)
            this.emitVotesChanged()
        }
    }

    startInteractionExecution(interaction: SuggestedInteraction) {
        this.ongoingInteractionExecution = interaction
        this.emit({ type: "interaction-execution-started", interaction })
    }

    finishInteractionExecution(id: string, requireStarted: boolean = true) {
        if (requireStarted && (this.ongoingInteractionExecution === null || id !== getSuggestedInteractionId(this.ongoingInteractionExecution))) return
        this.ongoingInteractionExecution = null
        this.emit({ type: "interaction-execution-finished" })
        this.interactionQueue.delete(id)
        this.emitVotesChanged()

        const interaction = this.definition.interactionsById.get(id)
        if (interaction === undefined) return

        switch (interaction.type) {
            case "use":
                if (interaction.consume) {
                    this.emit({ type: "object-visibility-changed", id: interaction.objectId, isVisible: false })
                }

                break

            case "combine":
                interaction.inputObjects.entries().forEach(([id, { consume }]) => {
                    if (consume) this.setObjectVisibility(id, false)
                })

                interaction.outputObjectIds.forEach(id => this.setObjectVisibility(id, true))

                break
        }
    }

    removeInteractionQueueItem(id: string) {
        if (!this.interactionQueue.has(id)) return
        this.interactionQueue.delete(id)
        this.emitVotesChanged()
    }

    setObjectVisibility(objectId: string, isVisible: boolean) {
        const current = this.objectVisibilityById.get(objectId)
        if (current === undefined) throw new Error(`Unknown object: ${objectId}`)
        if (current === isVisible) return

        this.objectVisibilityById.set(objectId, isVisible)
        this.emit({ type: "object-visibility-changed", id: objectId, isVisible })
    }
}